from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from eventstoredb.streams.types import StreamPosition, StreamRevision


@dataclass
class SubscribeToStreamOptions:
    from_revision: StreamPosition | StreamRevision = StreamPosition.START
    resolve_links: bool = False


@dataclass
class SubscriptionConfirmation:
    id: UUID



from __future__ import annotations

import json
from uuid import UUID

import betterproto

from eventstoredb.events import (
    BinaryRecordedEvent,
    ContentType,
    JsonRecordedEvent,
    Position,
    ReadEvent,
)
from eventstoredb.generated.event_store.client import Empty, StreamIdentifier
from eventstoredb.generated.event_store.client.streams import (
    ReadReqOptions,
    ReadReqOptionsReadDirection,
    ReadReqOptionsStreamOptions,
    ReadReqOptionsUuidOption,
    ReadResp,
    ReadRespReadEvent,
    ReadRespReadEventRecordedEvent,
)
from eventstoredb.streams.read.exceptions import StreamNotFoundError
from eventstoredb.streams.read.types import ReadDirection, ReadStreamOptions
from eventstoredb.streams.types import StreamPosition, StreamRevision


def create_read_request_options(
    stream_name: str,
    options: ReadStreamOptions | None = None,
) -> ReadReqOptions:
    if options is None:
        options = ReadStreamOptions()

    request_options = ReadReqOptions()
    request_options.resolve_links = options.resolve_links
    request_options.count = options.max_count
    request_options.no_filter = Empty()
    request_options.uuid_option = ReadReqOptionsUuidOption(string=Empty())

    if options.direction == ReadDirection.FORWARDS:
        request_options.read_direction = ReadReqOptionsReadDirection.Forwards
    elif options.direction == ReadDirection.BACKWARDS:
        request_options.read_direction = ReadReqOptionsReadDirection.Backwards

    request_options.stream = ReadReqOptionsStreamOptions()
    request_options.stream.stream_identifier = StreamIdentifier(stream_name.encode())

    if isinstance(options.from_revision, StreamRevision):
        request_options.stream.revision = options.from_revision
    elif options.from_revision == StreamPosition.START:
        request_options.stream.start = Empty()
    elif options.from_revision == StreamPosition.END:
        request_options.stream.end = Empty()

    return request_options


def convert_read_response(message: ReadResp) -> ReadEvent:
    content_type, _ = betterproto.which_one_of(message, "content")
    if content_type == "stream_not_found":
        raise StreamNotFoundError(
            stream_name=message.stream_not_found.stream_identifier.stream_name.decode()
        )
    elif content_type == "event":
        return convert_read_response_read_event(message.event)
    else:
        # FIXME maybe we should raise something like "UnexpectedRuntimeError" here?
        raise Exception("i shouldnt be here")


def convert_read_response_read_event(message: ReadRespReadEvent) -> ReadEvent:
    event = ReadEvent()
    if message.event:
        event.event = convert_read_response_recorded_event(message.event)
    if message.link:
        event.link = convert_read_response_recorded_event(message.link)
    # TODO should this use which_one_of?
    if message.commit_position:
        event.commit_position = message.commit_position
    return event


def convert_read_response_recorded_event(
    message: ReadRespReadEventRecordedEvent,
) -> JsonRecordedEvent | BinaryRecordedEvent:
    stream_name = message.stream_identifier.stream_name.decode()
    id = UUID(message.id.string)
    content_type = ContentType(message.metadata["content-type"])
    position = Position(
        commit=message.commit_position,
        prepare=message.prepare_position,
    )
    data = message.data
    metadata = message.custom_metadata
    # TODO reduce duplication
    if content_type == ContentType.JSON:
        data = json.loads(data.decode()) if data else None
        metadata = json.loads(metadata.decode()) if metadata else None
        return JsonRecordedEvent(
            stream_name=stream_name,
            id=id,
            revision=message.stream_revision,
            type=message.metadata["type"],
            content_type=content_type,
            created=int(message.metadata["created"]),
            position=position,
            data=data,
            metadata=metadata,
        )
    return BinaryRecordedEvent(
        stream_name=stream_name,
        id=id,
        revision=message.stream_revision,
        type=message.metadata["type"],
        content_type=content_type,
        created=int(message.metadata["created"]),
        position=position,
        data=data,
        metadata=metadata,
    )



from __future__ import annotations

from uuid import UUID

import betterproto

from eventstoredb.events import ReadEvent
from eventstoredb.generated.event_store.client import Empty, StreamIdentifier
from eventstoredb.generated.event_store.client.streams import (
    ReadReqOptions,
    ReadReqOptionsStreamOptions,
    ReadReqOptionsSubscriptionOptions,
    ReadReqOptionsUuidOption,
    ReadResp,
    ReadRespSubscriptionConfirmation,
)
from eventstoredb.streams.read.grpc import convert_read_response_read_event
from eventstoredb.streams.subscribe.types import (
    SubscribeToStreamOptions,
    SubscriptionConfirmation,
)
from eventstoredb.streams.types import StreamPosition, StreamRevision


def create_stream_subscription_options(
    stream_name: str,
    options: SubscribeToStreamOptions | None = None,
) -> ReadReqOptions:
    if options is None:
        options = SubscribeToStreamOptions()

    request_options = ReadReqOptions()
    request_options.resolve_links = options.resolve_links
    request_options.no_filter = Empty()
    request_options.uuid_option = ReadReqOptionsUuidOption(string=Empty())
    request_options.subscription = ReadReqOptionsSubscriptionOptions()

    request_options.stream = ReadReqOptionsStreamOptions()
    request_options.stream.stream_identifier = StreamIdentifier(stream_name.encode())

    if isinstance(options.from_revision, StreamRevision):
        request_options.stream.revision = options.from_revision
    elif options.from_revision == StreamPosition.START:
        request_options.stream.start = Empty()
    elif options.from_revision == StreamPosition.END:
        request_options.stream.end = Empty()

    return request_options


def convert_read_response(
    message: ReadResp,
) -> SubscriptionConfirmation | ReadEvent:
    content_type, _ = betterproto.which_one_of(message, "content")
    if content_type == "event":
        return convert_read_response_read_event(message.event)
    elif content_type == "confirmation":
        return convert_read_response_confirmation(message.confirmation)
    else:
        # TODO raise better exception
        raise Exception("shouldnt be here")


def convert_read_response_confirmation(
    message: ReadRespSubscriptionConfirmation,
) -> SubscriptionConfirmation:
    return SubscriptionConfirmation(id=UUID(message.subscription_id))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: cluster.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import Dict, List

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


class MemberInfoVNodeState(betterproto.Enum):
    Initializing = 0
    DiscoverLeader = 1
    Unknown = 2
    PreReplica = 3
    CatchingUp = 4
    Clone = 5
    Follower = 6
    PreLeader = 7
    Leader = 8
    Manager = 9
    ShuttingDown = 10
    Shutdown = 11
    ReadOnlyLeaderless = 12
    PreReadOnlyReplica = 13
    ReadOnlyReplica = 14
    ResigningLeader = 15


@dataclass(eq=False, repr=False)
class GossipRequest(betterproto.Message):
    info: "ClusterInfo" = betterproto.message_field(1)
    server: "EndPoint" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class ViewChangeRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    attempted_view: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class ViewChangeProofRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    installed_view: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class PrepareRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    view: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class PrepareOkRequest(betterproto.Message):
    view: int = betterproto.int32_field(1)
    server_id: "_client__.Uuid" = betterproto.message_field(2)
    server_http: "EndPoint" = betterproto.message_field(3)
    epoch_number: int = betterproto.int32_field(4)
    epoch_position: int = betterproto.int64_field(5)
    epoch_id: "_client__.Uuid" = betterproto.message_field(6)
    epoch_leader_instance_id: "_client__.Uuid" = betterproto.message_field(7)
    last_commit_position: int = betterproto.int64_field(8)
    writer_checkpoint: int = betterproto.int64_field(9)
    chaser_checkpoint: int = betterproto.int64_field(10)
    node_priority: int = betterproto.int32_field(11)
    cluster_info: "ClusterInfo" = betterproto.message_field(12)


@dataclass(eq=False, repr=False)
class ProposalRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    leader_id: "_client__.Uuid" = betterproto.message_field(3)
    leader_http: "EndPoint" = betterproto.message_field(4)
    view: int = betterproto.int32_field(5)
    epoch_number: int = betterproto.int32_field(6)
    epoch_position: int = betterproto.int64_field(7)
    epoch_id: "_client__.Uuid" = betterproto.message_field(8)
    epoch_leader_instance_id: "_client__.Uuid" = betterproto.message_field(9)
    last_commit_position: int = betterproto.int64_field(10)
    writer_checkpoint: int = betterproto.int64_field(11)
    chaser_checkpoint: int = betterproto.int64_field(12)
    node_priority: int = betterproto.int32_field(13)


@dataclass(eq=False, repr=False)
class AcceptRequest(betterproto.Message):
    server_id: "_client__.Uuid" = betterproto.message_field(1)
    server_http: "EndPoint" = betterproto.message_field(2)
    leader_id: "_client__.Uuid" = betterproto.message_field(3)
    leader_http: "EndPoint" = betterproto.message_field(4)
    view: int = betterproto.int32_field(5)


@dataclass(eq=False, repr=False)
class LeaderIsResigningRequest(betterproto.Message):
    leader_id: "_client__.Uuid" = betterproto.message_field(1)
    leader_http: "EndPoint" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class LeaderIsResigningOkRequest(betterproto.Message):
    leader_id: "_client__.Uuid" = betterproto.message_field(1)
    leader_http: "EndPoint" = betterproto.message_field(2)
    server_id: "_client__.Uuid" = betterproto.message_field(3)
    server_http: "EndPoint" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class ClusterInfo(betterproto.Message):
    members: List["MemberInfo"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class EndPoint(betterproto.Message):
    address: str = betterproto.string_field(1)
    port: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class MemberInfo(betterproto.Message):
    instance_id: "_client__.Uuid" = betterproto.message_field(1)
    time_stamp: int = betterproto.int64_field(2)
    state: "MemberInfoVNodeState" = betterproto.enum_field(3)
    is_alive: bool = betterproto.bool_field(4)
    http_end_point: "EndPoint" = betterproto.message_field(5)
    internal_tcp: "EndPoint" = betterproto.message_field(6)
    external_tcp: "EndPoint" = betterproto.message_field(7)
    internal_tcp_uses_tls: bool = betterproto.bool_field(8)
    external_tcp_uses_tls: bool = betterproto.bool_field(9)
    last_commit_position: int = betterproto.int64_field(10)
    writer_checkpoint: int = betterproto.int64_field(11)
    chaser_checkpoint: int = betterproto.int64_field(12)
    epoch_position: int = betterproto.int64_field(13)
    epoch_number: int = betterproto.int32_field(14)
    epoch_id: "_client__.Uuid" = betterproto.message_field(15)
    node_priority: int = betterproto.int32_field(16)
    is_read_only_replica: bool = betterproto.bool_field(17)
    advertise_host_to_client_as: str = betterproto.string_field(18)
    advertise_http_port_to_client_as: int = betterproto.uint32_field(19)
    advertise_tcp_port_to_client_as: int = betterproto.uint32_field(20)


class GossipStub(betterproto.ServiceStub):
    async def update(
        self, *, info: "ClusterInfo" = None, server: "EndPoint" = None
    ) -> "ClusterInfo":

        request = GossipRequest()
        if info is not None:
            request.info = info
        if server is not None:
            request.server = server

        return await self._unary_unary(
            "/event_store.cluster.Gossip/Update", request, ClusterInfo
        )

    async def read(self) -> "ClusterInfo":

        request = _client__.Empty()

        return await self._unary_unary(
            "/event_store.cluster.Gossip/Read", request, ClusterInfo
        )


class ElectionsStub(betterproto.ServiceStub):
    async def view_change(
        self,
        *,
        server_id: "_client__.Uuid" = None,
        server_http: "EndPoint" = None,
        attempted_view: int = 0
    ) -> "_client__.Empty":

        request = ViewChangeRequest()
        if server_id is not None:
            request.server_id = server_id
        if server_http is not None:
            request.server_http = server_http
        request.attempted_view = attempted_view

        return await self._unary_unary(
            "/event_store.cluster.Elections/ViewChange", request, _client__.Empty
        )

    async def view_change_proof(
        self,
        *,
        server_id: "_client__.Uuid" = None,
        server_http: "EndPoint" = None,
        installed_view: int = 0
    ) -> "_client__.Empty":

        request = ViewChangeProofRequest()
        if server_id is not None:
            request.server_id = server_id
        if server_http is not None:
            request.server_http = server_http
        request.installed_view = installed_view

        return await self._unary_unary(
            "/event_store.cluster.Elections/ViewChangeProof", request, _client__.Empty
        )

    async def prepare(
        self,
        *,
        server_id: "_client__.Uuid" = None,
        server_http: "EndPoint" = None,
        view: int = 0
    ) -> "_client__.Empty":

        request = PrepareRequest()
        if server_id is not None:
            request.server_id = server_id
        if server_http is not None:
            request.server_http = server_http
        request.view = view

        return await self._unary_unary(
            "/event_store.cluster.Elections/Prepare", request, _client__.Empty
        )

    async def prepare_ok(
        self,
        *,
        view: int = 0,
        server_id: "_client__.Uuid" = None,
        server_http: "EndPoint" = None,
        epoch_number: int = 0,
        epoch_position: int = 0,
        epoch_id: "_client__.Uuid" = None,
        epoch_leader_instance_id: "_client__.Uuid" = None,
        last_commit_position: int = 0,
        writer_checkpoint: int = 0,
        chaser_checkpoint: int = 0,
        node_priority: int = 0,
        cluster_info: "ClusterInfo" = None
    ) -> "_client__.Empty":

        request = PrepareOkRequest()
        request.view = view
        if server_id is not None:
            request.server_id = server_id
        if server_http is not None:
            request.server_http = server_http
        request.epoch_number = epoch_number
        request.epoch_position = epoch_position
        if epoch_id is not None:
            request.epoch_id = epoch_id
        if epoch_leader_instance_id is not None:
            request.epoch_leader_instance_id = epoch_leader_instance_id
        request.last_commit_position = last_commit_position
        request.writer_checkpoint = writer_checkpoint
        request.chaser_checkpoint = chaser_checkpoint
        request.node_priority = node_priority
        if cluster_info is not None:
            request.cluster_info = cluster_info

        return await self._unary_unary(
            "/event_store.cluster.Elections/PrepareOk", request, _client__.Empty
        )

    async def proposal(
        self,
        *,
        server_id: "_client__.Uuid" = None,
        server_http: "EndPoint" = None,
        leader_id: "_client__.Uuid" = None,
        leader_http: "EndPoint" = None,
        view: int = 0,
        epoch_number: int = 0,
        epoch_position: int = 0,
        epoch_id: "_client__.Uuid" = None,
        epoch_leader_instance_id: "_client__.Uuid" = None,
        last_commit_position: int = 0,
        writer_checkpoint: int = 0,
        chaser_checkpoint: int = 0,
        node_priority: int = 0
    ) -> "_client__.Empty":

        request = ProposalRequest()
        if server_id is not None:
            request.server_id = server_id
        if server_http is not None:
            request.server_http = server_http
        if leader_id is not None:
            request.leader_id = leader_id
        if leader_http is not None:
            request.leader_http = leader_http
        request.view = view
        request.epoch_number = epoch_number
        request.epoch_position = epoch_position
        if epoch_id is not None:
            request.epoch_id = epoch_id
        if epoch_leader_instance_id is not None:
            request.epoch_leader_instance_id = epoch_leader_instance_id
        request.last_commit_position = last_commit_position
        request.writer_checkpoint = writer_checkpoint
        request.chaser_checkpoint = chaser_checkpoint
        request.node_priority = node_priority

        return await self._unary_unary(
            "/event_store.cluster.Elections/Proposal", request, _client__.Empty
        )

    async def accept(
        self,
        *,
        server_id: "_client__.Uuid" = None,
        server_http: "EndPoint" = None,
        leader_id: "_client__.Uuid" = None,
        leader_http: "EndPoint" = None,
        view: int = 0
    ) -> "_client__.Empty":

        request = AcceptRequest()
        if server_id is not None:
            request.server_id = server_id
        if server_http is not None:
            request.server_http = server_http
        if leader_id is not None:
            request.leader_id = leader_id
        if leader_http is not None:
            request.leader_http = leader_http
        request.view = view

        return await self._unary_unary(
            "/event_store.cluster.Elections/Accept", request, _client__.Empty
        )

    async def leader_is_resigning(
        self, *, leader_id: "_client__.Uuid" = None, leader_http: "EndPoint" = None
    ) -> "_client__.Empty":

        request = LeaderIsResigningRequest()
        if leader_id is not None:
            request.leader_id = leader_id
        if leader_http is not None:
            request.leader_http = leader_http

        return await self._unary_unary(
            "/event_store.cluster.Elections/LeaderIsResigning", request, _client__.Empty
        )

    async def leader_is_resigning_ok(
        self,
        *,
        leader_id: "_client__.Uuid" = None,
        leader_http: "EndPoint" = None,
        server_id: "_client__.Uuid" = None,
        server_http: "EndPoint" = None
    ) -> "_client__.Empty":

        request = LeaderIsResigningOkRequest()
        if leader_id is not None:
            request.leader_id = leader_id
        if leader_http is not None:
            request.leader_http = leader_http
        if server_id is not None:
            request.server_id = server_id
        if server_http is not None:
            request.server_http = server_http

        return await self._unary_unary(
            "/event_store.cluster.Elections/LeaderIsResigningOk",
            request,
            _client__.Empty,
        )


class GossipBase(ServiceBase):
    async def update(self, info: "ClusterInfo", server: "EndPoint") -> "ClusterInfo":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def read(self) -> "ClusterInfo":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_update(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "info": request.info,
            "server": request.server,
        }

        response = await self.update(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_read(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {}

        response = await self.read(**request_kwargs)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.cluster.Gossip/Update": grpclib.const.Handler(
                self.__rpc_update,
                grpclib.const.Cardinality.UNARY_UNARY,
                GossipRequest,
                ClusterInfo,
            ),
            "/event_store.cluster.Gossip/Read": grpclib.const.Handler(
                self.__rpc_read,
                grpclib.const.Cardinality.UNARY_UNARY,
                _client__.Empty,
                ClusterInfo,
            ),
        }


class ElectionsBase(ServiceBase):
    async def view_change(
        self, server_id: "_client__.Uuid", server_http: "EndPoint", attempted_view: int
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def view_change_proof(
        self, server_id: "_client__.Uuid", server_http: "EndPoint", installed_view: int
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def prepare(
        self, server_id: "_client__.Uuid", server_http: "EndPoint", view: int
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def prepare_ok(
        self,
        view: int,
        server_id: "_client__.Uuid",
        server_http: "EndPoint",
        epoch_number: int,
        epoch_position: int,
        epoch_id: "_client__.Uuid",
        epoch_leader_instance_id: "_client__.Uuid",
        last_commit_position: int,
        writer_checkpoint: int,
        chaser_checkpoint: int,
        node_priority: int,
        cluster_info: "ClusterInfo",
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def proposal(
        self,
        server_id: "_client__.Uuid",
        server_http: "EndPoint",
        leader_id: "_client__.Uuid",
        leader_http: "EndPoint",
        view: int,
        epoch_number: int,
        epoch_position: int,
        epoch_id: "_client__.Uuid",
        epoch_leader_instance_id: "_client__.Uuid",
        last_commit_position: int,
        writer_checkpoint: int,
        chaser_checkpoint: int,
        node_priority: int,
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def accept(
        self,
        server_id: "_client__.Uuid",
        server_http: "EndPoint",
        leader_id: "_client__.Uuid",
        leader_http: "EndPoint",
        view: int,
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def leader_is_resigning(
        self, leader_id: "_client__.Uuid", leader_http: "EndPoint"
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def leader_is_resigning_ok(
        self,
        leader_id: "_client__.Uuid",
        leader_http: "EndPoint",
        server_id: "_client__.Uuid",
        server_http: "EndPoint",
    ) -> "_client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_view_change(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "server_id": request.server_id,
            "server_http": request.server_http,
            "attempted_view": request.attempted_view,
        }

        response = await self.view_change(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_view_change_proof(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "server_id": request.server_id,
            "server_http": request.server_http,
            "installed_view": request.installed_view,
        }

        response = await self.view_change_proof(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_prepare(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "server_id": request.server_id,
            "server_http": request.server_http,
            "view": request.view,
        }

        response = await self.prepare(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_prepare_ok(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "view": request.view,
            "server_id": request.server_id,
            "server_http": request.server_http,
            "epoch_number": request.epoch_number,
            "epoch_position": request.epoch_position,
            "epoch_id": request.epoch_id,
            "epoch_leader_instance_id": request.epoch_leader_instance_id,
            "last_commit_position": request.last_commit_position,
            "writer_checkpoint": request.writer_checkpoint,
            "chaser_checkpoint": request.chaser_checkpoint,
            "node_priority": request.node_priority,
            "cluster_info": request.cluster_info,
        }

        response = await self.prepare_ok(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_proposal(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "server_id": request.server_id,
            "server_http": request.server_http,
            "leader_id": request.leader_id,
            "leader_http": request.leader_http,
            "view": request.view,
            "epoch_number": request.epoch_number,
            "epoch_position": request.epoch_position,
            "epoch_id": request.epoch_id,
            "epoch_leader_instance_id": request.epoch_leader_instance_id,
            "last_commit_position": request.last_commit_position,
            "writer_checkpoint": request.writer_checkpoint,
            "chaser_checkpoint": request.chaser_checkpoint,
            "node_priority": request.node_priority,
        }

        response = await self.proposal(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_accept(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "server_id": request.server_id,
            "server_http": request.server_http,
            "leader_id": request.leader_id,
            "leader_http": request.leader_http,
            "view": request.view,
        }

        response = await self.accept(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_leader_is_resigning(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "leader_id": request.leader_id,
            "leader_http": request.leader_http,
        }

        response = await self.leader_is_resigning(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_leader_is_resigning_ok(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "leader_id": request.leader_id,
            "leader_http": request.leader_http,
            "server_id": request.server_id,
            "server_http": request.server_http,
        }

        response = await self.leader_is_resigning_ok(**request_kwargs)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.cluster.Elections/ViewChange": grpclib.const.Handler(
                self.__rpc_view_change,
                grpclib.const.Cardinality.UNARY_UNARY,
                ViewChangeRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/ViewChangeProof": grpclib.const.Handler(
                self.__rpc_view_change_proof,
                grpclib.const.Cardinality.UNARY_UNARY,
                ViewChangeProofRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/Prepare": grpclib.const.Handler(
                self.__rpc_prepare,
                grpclib.const.Cardinality.UNARY_UNARY,
                PrepareRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/PrepareOk": grpclib.const.Handler(
                self.__rpc_prepare_ok,
                grpclib.const.Cardinality.UNARY_UNARY,
                PrepareOkRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/Proposal": grpclib.const.Handler(
                self.__rpc_proposal,
                grpclib.const.Cardinality.UNARY_UNARY,
                ProposalRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/Accept": grpclib.const.Handler(
                self.__rpc_accept,
                grpclib.const.Cardinality.UNARY_UNARY,
                AcceptRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/LeaderIsResigning": grpclib.const.Handler(
                self.__rpc_leader_is_resigning,
                grpclib.const.Cardinality.UNARY_UNARY,
                LeaderIsResigningRequest,
                _client__.Empty,
            ),
            "/event_store.cluster.Elections/LeaderIsResigningOk": grpclib.const.Handler(
                self.__rpc_leader_is_resigning_ok,
                grpclib.const.Cardinality.UNARY_UNARY,
                LeaderIsResigningOkRequest,
                _client__.Empty,
            ),
        }


from .. import client as _client__



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: projections.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import AsyncIterator, Dict

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


@dataclass(eq=False, repr=False)
class CreateReq(betterproto.Message):
    options: "CreateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CreateReqOptions(betterproto.Message):
    one_time: "__client__.Empty" = betterproto.message_field(1, group="mode")
    transient: "CreateReqOptionsTransient" = betterproto.message_field(2, group="mode")
    continuous: "CreateReqOptionsContinuous" = betterproto.message_field(
        3, group="mode"
    )
    query: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class CreateReqOptionsTransient(betterproto.Message):
    name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CreateReqOptionsContinuous(betterproto.Message):
    name: str = betterproto.string_field(1)
    track_emitted_streams: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class CreateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class UpdateReq(betterproto.Message):
    options: "UpdateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class UpdateReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    query: str = betterproto.string_field(2)
    emit_enabled: bool = betterproto.bool_field(3, group="emit_option")
    no_emit_options: "__client__.Empty" = betterproto.message_field(
        4, group="emit_option"
    )


@dataclass(eq=False, repr=False)
class UpdateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DeleteReq(betterproto.Message):
    options: "DeleteReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DeleteReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    delete_emitted_streams: bool = betterproto.bool_field(2)
    delete_state_stream: bool = betterproto.bool_field(3)
    delete_checkpoint_stream: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class DeleteResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class StatisticsReq(betterproto.Message):
    options: "StatisticsReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class StatisticsReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1, group="mode")
    all: "__client__.Empty" = betterproto.message_field(2, group="mode")
    transient: "__client__.Empty" = betterproto.message_field(3, group="mode")
    continuous: "__client__.Empty" = betterproto.message_field(4, group="mode")
    one_time: "__client__.Empty" = betterproto.message_field(5, group="mode")


@dataclass(eq=False, repr=False)
class StatisticsResp(betterproto.Message):
    details: "StatisticsRespDetails" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class StatisticsRespDetails(betterproto.Message):
    core_processing_time: int = betterproto.int64_field(1)
    version: int = betterproto.int64_field(2)
    epoch: int = betterproto.int64_field(3)
    effective_name: str = betterproto.string_field(4)
    writes_in_progress: int = betterproto.int32_field(5)
    reads_in_progress: int = betterproto.int32_field(6)
    partitions_cached: int = betterproto.int32_field(7)
    status: str = betterproto.string_field(8)
    state_reason: str = betterproto.string_field(9)
    name: str = betterproto.string_field(10)
    mode: str = betterproto.string_field(11)
    position: str = betterproto.string_field(12)
    progress: float = betterproto.float_field(13)
    last_checkpoint: str = betterproto.string_field(14)
    events_processed_after_restart: int = betterproto.int64_field(15)
    checkpoint_status: str = betterproto.string_field(16)
    buffered_events: int = betterproto.int64_field(17)
    write_pending_events_before_checkpoint: int = betterproto.int32_field(18)
    write_pending_events_after_checkpoint: int = betterproto.int32_field(19)


@dataclass(eq=False, repr=False)
class StateReq(betterproto.Message):
    options: "StateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class StateReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    partition: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class StateResp(betterproto.Message):
    state: "betterproto_lib_google_protobuf.Value" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResultReq(betterproto.Message):
    options: "ResultReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResultReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    partition: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ResultResp(betterproto.Message):
    result: "betterproto_lib_google_protobuf.Value" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResetReq(betterproto.Message):
    options: "ResetReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResetReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    write_checkpoint: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class ResetResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class EnableReq(betterproto.Message):
    options: "EnableReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class EnableReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class EnableResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DisableReq(betterproto.Message):
    options: "DisableReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DisableReqOptions(betterproto.Message):
    name: str = betterproto.string_field(1)
    write_checkpoint: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class DisableResp(betterproto.Message):
    pass


class ProjectionsStub(betterproto.ServiceStub):
    async def create(self, *, options: "CreateReqOptions" = None) -> "CreateResp":

        request = CreateReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.projections.Projections/Create", request, CreateResp
        )

    async def update(self, *, options: "UpdateReqOptions" = None) -> "UpdateResp":

        request = UpdateReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.projections.Projections/Update", request, UpdateResp
        )

    async def delete(self, *, options: "DeleteReqOptions" = None) -> "DeleteResp":

        request = DeleteReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.projections.Projections/Delete", request, DeleteResp
        )

    async def statistics(
        self, *, options: "StatisticsReqOptions" = None
    ) -> AsyncIterator["StatisticsResp"]:

        request = StatisticsReq()
        if options is not None:
            request.options = options

        async for response in self._unary_stream(
            "/event_store.client.projections.Projections/Statistics",
            request,
            StatisticsResp,
        ):
            yield response

    async def disable(self, *, options: "DisableReqOptions" = None) -> "DisableResp":

        request = DisableReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.projections.Projections/Disable", request, DisableResp
        )

    async def enable(self, *, options: "EnableReqOptions" = None) -> "EnableResp":

        request = EnableReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.projections.Projections/Enable", request, EnableResp
        )

    async def reset(self, *, options: "ResetReqOptions" = None) -> "ResetResp":

        request = ResetReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.projections.Projections/Reset", request, ResetResp
        )

    async def state(self, *, options: "StateReqOptions" = None) -> "StateResp":

        request = StateReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.projections.Projections/State", request, StateResp
        )

    async def result(self, *, options: "ResultReqOptions" = None) -> "ResultResp":

        request = ResultReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.projections.Projections/Result", request, ResultResp
        )

    async def restart_subsystem(self) -> "__client__.Empty":

        request = __client__.Empty()

        return await self._unary_unary(
            "/event_store.client.projections.Projections/RestartSubsystem",
            request,
            __client__.Empty,
        )


class ProjectionsBase(ServiceBase):
    async def create(self, options: "CreateReqOptions") -> "CreateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def update(self, options: "UpdateReqOptions") -> "UpdateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete(self, options: "DeleteReqOptions") -> "DeleteResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def statistics(
        self, options: "StatisticsReqOptions"
    ) -> AsyncIterator["StatisticsResp"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def disable(self, options: "DisableReqOptions") -> "DisableResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def enable(self, options: "EnableReqOptions") -> "EnableResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def reset(self, options: "ResetReqOptions") -> "ResetResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def state(self, options: "StateReqOptions") -> "StateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def result(self, options: "ResultReqOptions") -> "ResultResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def restart_subsystem(self) -> "__client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_create(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.create(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_update(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.update(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_delete(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.delete(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_statistics(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        await self._call_rpc_handler_server_stream(
            self.statistics,
            stream,
            request_kwargs,
        )

    async def __rpc_disable(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.disable(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_enable(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.enable(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_reset(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.reset(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_state(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.state(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_result(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.result(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_restart_subsystem(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {}

        response = await self.restart_subsystem(**request_kwargs)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.projections.Projections/Create": grpclib.const.Handler(
                self.__rpc_create,
                grpclib.const.Cardinality.UNARY_UNARY,
                CreateReq,
                CreateResp,
            ),
            "/event_store.client.projections.Projections/Update": grpclib.const.Handler(
                self.__rpc_update,
                grpclib.const.Cardinality.UNARY_UNARY,
                UpdateReq,
                UpdateResp,
            ),
            "/event_store.client.projections.Projections/Delete": grpclib.const.Handler(
                self.__rpc_delete,
                grpclib.const.Cardinality.UNARY_UNARY,
                DeleteReq,
                DeleteResp,
            ),
            "/event_store.client.projections.Projections/Statistics": grpclib.const.Handler(
                self.__rpc_statistics,
                grpclib.const.Cardinality.UNARY_STREAM,
                StatisticsReq,
                StatisticsResp,
            ),
            "/event_store.client.projections.Projections/Disable": grpclib.const.Handler(
                self.__rpc_disable,
                grpclib.const.Cardinality.UNARY_UNARY,
                DisableReq,
                DisableResp,
            ),
            "/event_store.client.projections.Projections/Enable": grpclib.const.Handler(
                self.__rpc_enable,
                grpclib.const.Cardinality.UNARY_UNARY,
                EnableReq,
                EnableResp,
            ),
            "/event_store.client.projections.Projections/Reset": grpclib.const.Handler(
                self.__rpc_reset,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResetReq,
                ResetResp,
            ),
            "/event_store.client.projections.Projections/State": grpclib.const.Handler(
                self.__rpc_state,
                grpclib.const.Cardinality.UNARY_UNARY,
                StateReq,
                StateResp,
            ),
            "/event_store.client.projections.Projections/Result": grpclib.const.Handler(
                self.__rpc_result,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResultReq,
                ResultResp,
            ),
            "/event_store.client.projections.Projections/RestartSubsystem": grpclib.const.Handler(
                self.__rpc_restart_subsystem,
                grpclib.const.Cardinality.UNARY_UNARY,
                __client__.Empty,
                __client__.Empty,
            ),
        }


import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf

from ... import client as __client__



from eventstoredb.streams.read.exceptions import ReadStreamError, StreamNotFoundError
from eventstoredb.streams.read.grpc import (
    convert_read_response,
    create_read_request_options,
)
from eventstoredb.streams.read.types import ReadDirection, ReadStreamOptions



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: code.proto, status.proto
# plugin: python-betterproto
from dataclasses import dataclass

import betterproto
from betterproto.grpc.grpclib_server import ServiceBase


class Code(betterproto.Enum):
    """
    The canonical error codes for gRPC APIs. Sometimes multiple error codes may
    apply.  Services should return the most specific error code that applies.
    For example, prefer `OUT_OF_RANGE` over `FAILED_PRECONDITION` if both codes
    apply. Similarly prefer `NOT_FOUND` or `ALREADY_EXISTS` over
    `FAILED_PRECONDITION`.
    """

    # Not an error; returned on success HTTP Mapping: 200 OK
    OK = 0
    # The operation was cancelled, typically by the caller. HTTP Mapping: 499
    # Client Closed Request
    CANCELLED = 1
    # Unknown error.  For example, this error may be returned when a `Status`
    # value received from another address space belongs to an error space that is
    # not known in this address space.  Also errors raised by APIs that do not
    # return enough error information may be converted to this error. HTTP
    # Mapping: 500 Internal Server Error
    UNKNOWN = 2
    # The client specified an invalid argument.  Note that this differs from
    # `FAILED_PRECONDITION`.  `INVALID_ARGUMENT` indicates arguments that are
    # problematic regardless of the state of the system (e.g., a malformed file
    # name). HTTP Mapping: 400 Bad Request
    INVALID_ARGUMENT = 3
    # The deadline expired before the operation could complete. For operations
    # that change the state of the system, this error may be returned even if the
    # operation has completed successfully.  For example, a successful response
    # from a server could have been delayed long enough for the deadline to
    # expire. HTTP Mapping: 504 Gateway Timeout
    DEADLINE_EXCEEDED = 4
    # Some requested entity (e.g., file or directory) was not found. Note to
    # server developers: if a request is denied for an entire class of users,
    # such as gradual feature rollout or undocumented whitelist, `NOT_FOUND` may
    # be used. If a request is denied for some users within a class of users,
    # such as user-based access control, `PERMISSION_DENIED` must be used. HTTP
    # Mapping: 404 Not Found
    NOT_FOUND = 5
    # The entity that a client attempted to create (e.g., file or directory)
    # already exists. HTTP Mapping: 409 Conflict
    ALREADY_EXISTS = 6
    # The caller does not have permission to execute the specified operation.
    # `PERMISSION_DENIED` must not be used for rejections caused by exhausting
    # some resource (use `RESOURCE_EXHAUSTED` instead for those errors).
    # `PERMISSION_DENIED` must not be used if the caller can not be identified
    # (use `UNAUTHENTICATED` instead for those errors). This error code does not
    # imply the request is valid or the requested entity exists or satisfies
    # other pre-conditions. HTTP Mapping: 403 Forbidden
    PERMISSION_DENIED = 7
    # The request does not have valid authentication credentials for the
    # operation. HTTP Mapping: 401 Unauthorized
    UNAUTHENTICATED = 16
    # Some resource has been exhausted, perhaps a per-user quota, or perhaps the
    # entire file system is out of space. HTTP Mapping: 429 Too Many Requests
    RESOURCE_EXHAUSTED = 8
    # The operation was rejected because the system is not in a state required
    # for the operation's execution.  For example, the directory to be deleted is
    # non-empty, an rmdir operation is applied to a non-directory, etc. Service
    # implementors can use the following guidelines to decide between
    # `FAILED_PRECONDITION`, `ABORTED`, and `UNAVAILABLE`:  (a) Use `UNAVAILABLE`
    # if the client can retry just the failing call.  (b) Use `ABORTED` if the
    # client should retry at a higher level      (e.g., when a client-specified
    # test-and-set fails, indicating the      client should restart a read-
    # modify-write sequence).  (c) Use `FAILED_PRECONDITION` if the client should
    # not retry until      the system state has been explicitly fixed.  E.g., if
    # an "rmdir"      fails because the directory is non-empty,
    # `FAILED_PRECONDITION`      should be returned since the client should not
    # retry unless      the files are deleted from the directory. HTTP Mapping:
    # 400 Bad Request
    FAILED_PRECONDITION = 9
    # The operation was aborted, typically due to a concurrency issue such as a
    # sequencer check failure or transaction abort. See the guidelines above for
    # deciding between `FAILED_PRECONDITION`, `ABORTED`, and `UNAVAILABLE`. HTTP
    # Mapping: 409 Conflict
    ABORTED = 10
    # The operation was attempted past the valid range.  E.g., seeking or reading
    # past end-of-file. Unlike `INVALID_ARGUMENT`, this error indicates a problem
    # that may be fixed if the system state changes. For example, a 32-bit file
    # system will generate `INVALID_ARGUMENT` if asked to read at an offset that
    # is not in the range [0,2^32-1], but it will generate `OUT_OF_RANGE` if
    # asked to read from an offset past the current file size. There is a fair
    # bit of overlap between `FAILED_PRECONDITION` and `OUT_OF_RANGE`.  We
    # recommend using `OUT_OF_RANGE` (the more specific error) when it applies so
    # that callers who are iterating through a space can easily look for an
    # `OUT_OF_RANGE` error to detect when they are done. HTTP Mapping: 400 Bad
    # Request
    OUT_OF_RANGE = 11
    # The operation is not implemented or is not supported/enabled in this
    # service. HTTP Mapping: 501 Not Implemented
    UNIMPLEMENTED = 12
    # Internal errors.  This means that some invariants expected by the
    # underlying system have been broken.  This error code is reserved for
    # serious errors. HTTP Mapping: 500 Internal Server Error
    INTERNAL = 13
    # The service is currently unavailable.  This is most likely a transient
    # condition, which can be corrected by retrying with a backoff. Note that it
    # is not always safe to retry non-idempotent operations. See the guidelines
    # above for deciding between `FAILED_PRECONDITION`, `ABORTED`, and
    # `UNAVAILABLE`. HTTP Mapping: 503 Service Unavailable
    UNAVAILABLE = 14
    # Unrecoverable data loss or corruption. HTTP Mapping: 500 Internal Server
    # Error
    DATA_LOSS = 15


@dataclass(eq=False, repr=False)
class Status(betterproto.Message):
    """
    The `Status` type defines a logical error model that is suitable for
    different programming environments, including REST APIs and RPC APIs. It is
    used by [gRPC](https://github.com/grpc). Each `Status` message contains
    three pieces of data: error code, error message, and error details. You can
    find out more about this error model and how to work with it in the [API
    Design Guide](https://cloud.google.com/apis/design/errors).
    """

    # The status code, which should be an enum value of
    # [google.rpc.Code][google.rpc.Code].
    code: "Code" = betterproto.enum_field(1)
    # A developer-facing error message, which should be in English. Any user-
    # facing error message should be localized and sent in the
    # [google.rpc.Status.details][google.rpc.Status.details] field, or localized
    # by the client.
    message: str = betterproto.string_field(2)
    # A list of messages that carry the error details.  There is a common set of
    # message types for APIs to use.
    details: "betterproto_lib_google_protobuf.Any" = betterproto.message_field(3)


import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: monitoring.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import AsyncIterator, Dict

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


@dataclass(eq=False, repr=False)
class StatsReq(betterproto.Message):
    use_metadata: bool = betterproto.bool_field(1)
    refresh_time_period_in_ms: int = betterproto.uint64_field(4)


@dataclass(eq=False, repr=False)
class StatsResp(betterproto.Message):
    stats: Dict[str, str] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )


class MonitoringStub(betterproto.ServiceStub):
    async def stats(
        self, *, use_metadata: bool = False, refresh_time_period_in_ms: int = 0
    ) -> AsyncIterator["StatsResp"]:

        request = StatsReq()
        request.use_metadata = use_metadata
        request.refresh_time_period_in_ms = refresh_time_period_in_ms

        async for response in self._unary_stream(
            "/event_store.client.monitoring.Monitoring/Stats",
            request,
            StatsResp,
        ):
            yield response


class MonitoringBase(ServiceBase):
    async def stats(
        self, use_metadata: bool, refresh_time_period_in_ms: int
    ) -> AsyncIterator["StatsResp"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_stats(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "use_metadata": request.use_metadata,
            "refresh_time_period_in_ms": request.refresh_time_period_in_ms,
        }

        await self._call_rpc_handler_server_stream(
            self.stats,
            stream,
            request_kwargs,
        )

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.monitoring.Monitoring/Stats": grpclib.const.Handler(
                self.__rpc_stats,
                grpclib.const.Cardinality.UNARY_STREAM,
                StatsReq,
                StatsResp,
            ),
        }



from eventstoredb.streams.subscribe.grpc import create_stream_subscription_options
from eventstoredb.streams.subscribe.subscription import Subscription
from eventstoredb.streams.subscribe.types import SubscribeToStreamOptions



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: serverfeatures.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import Dict, List

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


@dataclass(eq=False, repr=False)
class SupportedMethods(betterproto.Message):
    methods: List["SupportedMethod"] = betterproto.message_field(1)
    event_store_server_version: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class SupportedMethod(betterproto.Message):
    method_name: str = betterproto.string_field(1)
    service_name: str = betterproto.string_field(2)
    features: List[str] = betterproto.string_field(3)


class ServerFeaturesStub(betterproto.ServiceStub):
    async def get_supported_methods(self) -> "SupportedMethods":

        request = __client__.Empty()

        return await self._unary_unary(
            "/event_store.client.server_features.ServerFeatures/GetSupportedMethods",
            request,
            SupportedMethods,
        )


class ServerFeaturesBase(ServiceBase):
    async def get_supported_methods(self) -> "SupportedMethods":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_get_supported_methods(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {}

        response = await self.get_supported_methods(**request_kwargs)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.server_features.ServerFeatures/GetSupportedMethods": grpclib.const.Handler(
                self.__rpc_get_supported_methods,
                grpclib.const.Cardinality.UNARY_UNARY,
                __client__.Empty,
                SupportedMethods,
            ),
        }


from ... import client as __client__



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: operations.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import Dict

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


class ScavengeRespScavengeResult(betterproto.Enum):
    Started = 0
    InProgress = 1
    Stopped = 2


@dataclass(eq=False, repr=False)
class StartScavengeReq(betterproto.Message):
    options: "StartScavengeReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class StartScavengeReqOptions(betterproto.Message):
    thread_count: int = betterproto.int32_field(1)
    start_from_chunk: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class StopScavengeReq(betterproto.Message):
    options: "StopScavengeReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class StopScavengeReqOptions(betterproto.Message):
    scavenge_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ScavengeResp(betterproto.Message):
    scavenge_id: str = betterproto.string_field(1)
    scavenge_result: "ScavengeRespScavengeResult" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class SetNodePriorityReq(betterproto.Message):
    priority: int = betterproto.int32_field(1)


class OperationsStub(betterproto.ServiceStub):
    async def start_scavenge(
        self, *, options: "StartScavengeReqOptions" = None
    ) -> "ScavengeResp":

        request = StartScavengeReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.operations.Operations/StartScavenge",
            request,
            ScavengeResp,
        )

    async def stop_scavenge(
        self, *, options: "StopScavengeReqOptions" = None
    ) -> "ScavengeResp":

        request = StopScavengeReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.operations.Operations/StopScavenge",
            request,
            ScavengeResp,
        )

    async def shutdown(self) -> "__client__.Empty":

        request = __client__.Empty()

        return await self._unary_unary(
            "/event_store.client.operations.Operations/Shutdown",
            request,
            __client__.Empty,
        )

    async def merge_indexes(self) -> "__client__.Empty":

        request = __client__.Empty()

        return await self._unary_unary(
            "/event_store.client.operations.Operations/MergeIndexes",
            request,
            __client__.Empty,
        )

    async def resign_node(self) -> "__client__.Empty":

        request = __client__.Empty()

        return await self._unary_unary(
            "/event_store.client.operations.Operations/ResignNode",
            request,
            __client__.Empty,
        )

    async def set_node_priority(self, *, priority: int = 0) -> "__client__.Empty":

        request = SetNodePriorityReq()
        request.priority = priority

        return await self._unary_unary(
            "/event_store.client.operations.Operations/SetNodePriority",
            request,
            __client__.Empty,
        )

    async def restart_persistent_subscriptions(self) -> "__client__.Empty":

        request = __client__.Empty()

        return await self._unary_unary(
            "/event_store.client.operations.Operations/RestartPersistentSubscriptions",
            request,
            __client__.Empty,
        )


class OperationsBase(ServiceBase):
    async def start_scavenge(
        self, options: "StartScavengeReqOptions"
    ) -> "ScavengeResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def stop_scavenge(self, options: "StopScavengeReqOptions") -> "ScavengeResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def shutdown(self) -> "__client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def merge_indexes(self) -> "__client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def resign_node(self) -> "__client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def set_node_priority(self, priority: int) -> "__client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def restart_persistent_subscriptions(self) -> "__client__.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_start_scavenge(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.start_scavenge(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_stop_scavenge(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.stop_scavenge(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_shutdown(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {}

        response = await self.shutdown(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_merge_indexes(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {}

        response = await self.merge_indexes(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_resign_node(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {}

        response = await self.resign_node(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_set_node_priority(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "priority": request.priority,
        }

        response = await self.set_node_priority(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_restart_persistent_subscriptions(
        self, stream: grpclib.server.Stream
    ) -> None:
        request = await stream.recv_message()

        request_kwargs = {}

        response = await self.restart_persistent_subscriptions(**request_kwargs)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.operations.Operations/StartScavenge": grpclib.const.Handler(
                self.__rpc_start_scavenge,
                grpclib.const.Cardinality.UNARY_UNARY,
                StartScavengeReq,
                ScavengeResp,
            ),
            "/event_store.client.operations.Operations/StopScavenge": grpclib.const.Handler(
                self.__rpc_stop_scavenge,
                grpclib.const.Cardinality.UNARY_UNARY,
                StopScavengeReq,
                ScavengeResp,
            ),
            "/event_store.client.operations.Operations/Shutdown": grpclib.const.Handler(
                self.__rpc_shutdown,
                grpclib.const.Cardinality.UNARY_UNARY,
                __client__.Empty,
                __client__.Empty,
            ),
            "/event_store.client.operations.Operations/MergeIndexes": grpclib.const.Handler(
                self.__rpc_merge_indexes,
                grpclib.const.Cardinality.UNARY_UNARY,
                __client__.Empty,
                __client__.Empty,
            ),
            "/event_store.client.operations.Operations/ResignNode": grpclib.const.Handler(
                self.__rpc_resign_node,
                grpclib.const.Cardinality.UNARY_UNARY,
                __client__.Empty,
                __client__.Empty,
            ),
            "/event_store.client.operations.Operations/SetNodePriority": grpclib.const.Handler(
                self.__rpc_set_node_priority,
                grpclib.const.Cardinality.UNARY_UNARY,
                SetNodePriorityReq,
                __client__.Empty,
            ),
            "/event_store.client.operations.Operations/RestartPersistentSubscriptions": grpclib.const.Handler(
                self.__rpc_restart_persistent_subscriptions,
                grpclib.const.Cardinality.UNARY_UNARY,
                __client__.Empty,
                __client__.Empty,
            ),
        }


from ... import client as __client__



from __future__ import annotations

import json
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Literal
from uuid import UUID, uuid4


@dataclass
class Position:
    commit: int
    prepare: int


class ContentType(str, Enum):
    BINARY = "application/octet-stream"
    JSON = "application/json"


@dataclass
class EventData:
    # NOTE order matters due to dataclass-inheritence
    type: str
    content_type: ContentType
    id: UUID = field(default_factory=uuid4)
    data: bytes | None = None
    metadata: bytes | None = None

    def _serialized_data(self) -> bytes | None:
        return self.data

    def _serialized_metadata(self) -> bytes | None:
        return self.metadata


@dataclass
class JsonEvent(EventData):
    content_type: Literal[ContentType.JSON] = ContentType.JSON
    # TODO it would be great not to use Any here
    data: Any | None = None
    # TODO it would be great not to use Any here
    metadata: Any | None = None

    def _serialized_data(self) -> bytes | None:
        if self.data:
            json_data = json.dumps(self.data)
            return json_data.encode()
        return None

    def _serialized_metadata(self) -> bytes | None:
        if self.metadata:
            json_data = json.dumps(self.metadata)
            return json_data.encode()
        return None


@dataclass
class BinaryEvent(EventData):
    content_type: Literal[ContentType.BINARY] = ContentType.BINARY


@dataclass
class RecordedEvent:
    stream_name: str
    id: UUID
    type: str
    content_type: ContentType
    revision: int
    created: int
    position: Position


@dataclass
class JsonRecordedEvent(RecordedEvent):
    # TODO it would be great not to use Any here
    data: Any | None
    # TODO it would be great not to use Any here
    metadata: Any | None


@dataclass
class BinaryRecordedEvent(RecordedEvent):
    data: bytes | None
    metadata: bytes | None


# TODO does this belong to streams/read/types?
@dataclass
class ReadEvent:
    event: RecordedEvent | None = None
    link: RecordedEvent | None = None
    commit_position: int | None = None



from eventstoredb.persistent_subscriptions.common.types import (
    PersistentSubscriptionSettings,
)
from eventstoredb.persistent_subscriptions.update.grpc import (
    create_update_request_options,
)
from eventstoredb.persistent_subscriptions.update.types import (
    UpdatePersistentSubscriptionOptions,
)



from __future__ import annotations

from typing import AsyncIterator
from uuid import UUID

from eventstoredb.events import ReadEvent
from eventstoredb.generated.event_store.client.streams import ReadResp
from eventstoredb.streams.subscribe.grpc import convert_read_response
from eventstoredb.streams.subscribe.types import SubscriptionConfirmation


class Subscription(AsyncIterator[ReadEvent]):
    def __init__(self, it: AsyncIterator[ReadResp]) -> None:
        self._it = it
        self.id: UUID | None

    def __aiter__(self) -> AsyncIterator[ReadEvent]:
        return self

    async def __anext__(self) -> ReadEvent:
        while True:
            response = await self._it.__anext__()
            event = convert_read_response(response)
            if isinstance(event, SubscriptionConfirmation):
                self.id = event.id
            else:
                return event



from eventstoredb.persistent_subscriptions.delete.grpc import (
    create_delete_request_options,
)
from eventstoredb.persistent_subscriptions.delete.types import (
    DeletePersistentSubscriptionOptions,
)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: users.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import AsyncIterator, Dict, List

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase


@dataclass(eq=False, repr=False)
class CreateReq(betterproto.Message):
    options: "CreateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CreateReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    password: str = betterproto.string_field(2)
    full_name: str = betterproto.string_field(3)
    groups: List[str] = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class CreateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class UpdateReq(betterproto.Message):
    options: "UpdateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class UpdateReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    password: str = betterproto.string_field(2)
    full_name: str = betterproto.string_field(3)
    groups: List[str] = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class UpdateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DeleteReq(betterproto.Message):
    options: "DeleteReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DeleteReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class DeleteResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class EnableReq(betterproto.Message):
    options: "EnableReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class EnableReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class EnableResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DisableReq(betterproto.Message):
    options: "DisableReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DisableReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class DisableResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DetailsReq(betterproto.Message):
    options: "DetailsReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DetailsReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class DetailsResp(betterproto.Message):
    user_details: "DetailsRespUserDetails" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DetailsRespUserDetails(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    full_name: str = betterproto.string_field(2)
    groups: List[str] = betterproto.string_field(3)
    last_updated: "DetailsRespUserDetailsDateTime" = betterproto.message_field(4)
    disabled: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class DetailsRespUserDetailsDateTime(betterproto.Message):
    ticks_since_epoch: int = betterproto.int64_field(1)


@dataclass(eq=False, repr=False)
class ChangePasswordReq(betterproto.Message):
    options: "ChangePasswordReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ChangePasswordReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    current_password: str = betterproto.string_field(2)
    new_password: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class ChangePasswordResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ResetPasswordReq(betterproto.Message):
    options: "ResetPasswordReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResetPasswordReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    new_password: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ResetPasswordResp(betterproto.Message):
    pass


class UsersStub(betterproto.ServiceStub):
    async def create(self, *, options: "CreateReqOptions" = None) -> "CreateResp":

        request = CreateReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.users.Users/Create", request, CreateResp
        )

    async def update(self, *, options: "UpdateReqOptions" = None) -> "UpdateResp":

        request = UpdateReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.users.Users/Update", request, UpdateResp
        )

    async def delete(self, *, options: "DeleteReqOptions" = None) -> "DeleteResp":

        request = DeleteReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.users.Users/Delete", request, DeleteResp
        )

    async def disable(self, *, options: "DisableReqOptions" = None) -> "DisableResp":

        request = DisableReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.users.Users/Disable", request, DisableResp
        )

    async def enable(self, *, options: "EnableReqOptions" = None) -> "EnableResp":

        request = EnableReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.users.Users/Enable", request, EnableResp
        )

    async def details(
        self, *, options: "DetailsReqOptions" = None
    ) -> AsyncIterator["DetailsResp"]:

        request = DetailsReq()
        if options is not None:
            request.options = options

        async for response in self._unary_stream(
            "/event_store.client.users.Users/Details",
            request,
            DetailsResp,
        ):
            yield response

    async def change_password(
        self, *, options: "ChangePasswordReqOptions" = None
    ) -> "ChangePasswordResp":

        request = ChangePasswordReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.users.Users/ChangePassword",
            request,
            ChangePasswordResp,
        )

    async def reset_password(
        self, *, options: "ResetPasswordReqOptions" = None
    ) -> "ResetPasswordResp":

        request = ResetPasswordReq()
        if options is not None:
            request.options = options

        return await self._unary_unary(
            "/event_store.client.users.Users/ResetPassword", request, ResetPasswordResp
        )


class UsersBase(ServiceBase):
    async def create(self, options: "CreateReqOptions") -> "CreateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def update(self, options: "UpdateReqOptions") -> "UpdateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete(self, options: "DeleteReqOptions") -> "DeleteResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def disable(self, options: "DisableReqOptions") -> "DisableResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def enable(self, options: "EnableReqOptions") -> "EnableResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def details(
        self, options: "DetailsReqOptions"
    ) -> AsyncIterator["DetailsResp"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def change_password(
        self, options: "ChangePasswordReqOptions"
    ) -> "ChangePasswordResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def reset_password(
        self, options: "ResetPasswordReqOptions"
    ) -> "ResetPasswordResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_create(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.create(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_update(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.update(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_delete(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.delete(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_disable(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.disable(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_enable(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.enable(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_details(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        await self._call_rpc_handler_server_stream(
            self.details,
            stream,
            request_kwargs,
        )

    async def __rpc_change_password(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.change_password(**request_kwargs)
        await stream.send_message(response)

    async def __rpc_reset_password(self, stream: grpclib.server.Stream) -> None:
        request = await stream.recv_message()

        request_kwargs = {
            "options": request.options,
        }

        response = await self.reset_password(**request_kwargs)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.users.Users/Create": grpclib.const.Handler(
                self.__rpc_create,
                grpclib.const.Cardinality.UNARY_UNARY,
                CreateReq,
                CreateResp,
            ),
            "/event_store.client.users.Users/Update": grpclib.const.Handler(
                self.__rpc_update,
                grpclib.const.Cardinality.UNARY_UNARY,
                UpdateReq,
                UpdateResp,
            ),
            "/event_store.client.users.Users/Delete": grpclib.const.Handler(
                self.__rpc_delete,
                grpclib.const.Cardinality.UNARY_UNARY,
                DeleteReq,
                DeleteResp,
            ),
            "/event_store.client.users.Users/Disable": grpclib.const.Handler(
                self.__rpc_disable,
                grpclib.const.Cardinality.UNARY_UNARY,
                DisableReq,
                DisableResp,
            ),
            "/event_store.client.users.Users/Enable": grpclib.const.Handler(
                self.__rpc_enable,
                grpclib.const.Cardinality.UNARY_UNARY,
                EnableReq,
                EnableResp,
            ),
            "/event_store.client.users.Users/Details": grpclib.const.Handler(
                self.__rpc_details,
                grpclib.const.Cardinality.UNARY_STREAM,
                DetailsReq,
                DetailsResp,
            ),
            "/event_store.client.users.Users/ChangePassword": grpclib.const.Handler(
                self.__rpc_change_password,
                grpclib.const.Cardinality.UNARY_UNARY,
                ChangePasswordReq,
                ChangePasswordResp,
            ),
            "/event_store.client.users.Users/ResetPassword": grpclib.const.Handler(
                self.__rpc_reset_password,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResetPasswordReq,
                ResetPasswordResp,
            ),
        }
